#lang plai-typed
(require plai-typed/s-exp-match
         "class.rkt"
         "inherit.rkt")

(module+ test
  (print-only-errors true))

;; ----------------------------------------

(define (parse-class [s : s-expression]) : (ClassI 't)
  (cond
   [(s-exp-match? `{class SYMBOL extends SYMBOL {ANY ...} ANY ...} s)
    (classI (s-exp->symbol (second (s-exp->list s)))
            (s-exp->symbol (fourth (s-exp->list s)))
            (map parse-field
                 (s-exp->list (fourth (rest (s-exp->list s)))))
            (map parse-method
                 (rest (rest (rest (rest (rest (s-exp->list s))))))))]
   [else (error 'parse-class "invalid input")]))

(define (parse-field [s : s-expression]) : symbol
  (cond
   [(s-exp-match? `SYMBOL s)
    (s-exp->symbol s)]
   [else (error 'parse-field "invalid input")]))

(define (parse-method [s : s-expression]) : (MethodI 't)
  (cond
   [(s-exp-match? `{SYMBOL ANY} s)
    (methodI (s-exp->symbol (first (s-exp->list s)))
             (parse (second (s-exp->list s)) box))]
   [else (error 'parse-method "invalid input")]))

(define (parse [s : s-expression]
               [parse-type : (s-expression -> 't)]) : (ExprI 't)
  (cond
   [(s-exp-match? `NUMBER s) (numI (s-exp->number s))]
   [(s-exp-match? `arg s) (argI)]
   [(s-exp-match? `this s) (thisI)]
   [(s-exp-match? '{- ANY ANY} s)
    (subI (parse (second (s-exp->list s)) parse-type)
          (parse (third (s-exp->list s)) parse-type))]
   [(s-exp-match? '{+ ANY ANY} s)
    (plusI (parse (second (s-exp->list s)) parse-type)
           (parse (third (s-exp->list s)) parse-type))]
   [(s-exp-match? '{* ANY ANY} s)
    (multI (parse (second (s-exp->list s)) parse-type)
           (parse (third (s-exp->list s)) parse-type))]
   [(s-exp-match? '{box ANY ANY} s)
    (boxI (parse-type (second (s-exp->list s)))
          (parse (third (s-exp->list s)) parse-type))]
   [(s-exp-match? '{unbox ANY} s)
    (unboxI (parse (second (s-exp->list s)) parse-type))]
   [(s-exp-match? '{set-box! ANY ANY} s)
    (set-boxI (parse (second (s-exp->list s)) parse-type)
              (parse (third (s-exp->list s)) parse-type))]
   [(s-exp-match? '{new SYMBOL ANY ...} s)
    (newI (s-exp->symbol (second (s-exp->list s)))
          (map (lambda (s) (parse s parse-type))
               (rest (rest (s-exp->list s)))))]
   [(s-exp-match? '{get ANY SYMBOL} s)
    (getI (parse (second (s-exp->list s)) parse-type)
          (s-exp->symbol (third (s-exp->list s))))]
   [(s-exp-match? '{send ANY SYMBOL ANY} s)
    (sendI (parse (second (s-exp->list s)) parse-type)
           (s-exp->symbol (third (s-exp->list s)))
           (parse (fourth (s-exp->list s)) parse-type))]
   [(s-exp-match? '{super SYMBOL ANY} s)
    (superI (s-exp->symbol (second (s-exp->list s)))
            (parse (third (s-exp->list s)) parse-type))]
   [else (error 'parse "invalid input")]))

(module+ test
  ;; These tests arbitrarily use `box` as the "parser"
  ;; for types.
  (test (parse '{- 1 2} box)
        (subI (numI 1) (numI 2)))
  (test (parse '0 box)
        (numI 0))
  (test (parse `arg box)
        (argI))
  (test (parse `this box)
        (thisI))
  (test (parse '{+ 1 2} box)
        (plusI (numI 1) (numI 2)))
  (test (parse '{* 1 2} box)
        (multI (numI 1) (numI 2)))
  (test (parse '{box x 0} box)
        (boxI (box `x) (numI 0)))
  (test (parse '{unbox 1} box)
        (unboxI (numI 1)))
  (test (parse '{set-box! 1 0} box)
        (set-boxI (numI 1) (numI 0)))
  (test (parse '{new posn 1 2} box)
        (newI 'posn (list (numI 1) (numI 2))))
  (test (parse '{get 1 x} box)
        (getI (numI 1) 'x))
  (test (parse '{send 1 m 2} box)
        (sendI (numI 1) 'm (numI 2)))
  (test (parse '{super m 1} box)
        (superI 'm (numI 1)))
  (test/exn (parse `x box)
            "invalid input")

  (test (parse-field `x)
        'x)
  (test/exn (parse-field '{x 1})
            "invalid input")

  (test (parse-method `{m this})
        (methodI 'm (thisI)))
  (test/exn (parse-method `{m 1 2})
            "invalid input")
  
  (test (parse-class '{class posn3D extends posn
                             {x y z}
                             {m1 arg}
                             {m2 this}})
        (classI 'posn3D 'posn
                (list 'x 'y 'z)
                (list (methodI 'm1 (argI))
                      (methodI 'm2 (thisI)))))
  (test/exn (parse-class '{class})
            "invalid input"))

